#import "MainController.h"
#import "main.h"

@implementation MainController

- (void)awakeFromNib
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	txtColor = [char01 backgroundColor];
	charList = [[NSMutableArray alloc] initWithCapacity:100];
	
	words   = [[NSMutableDictionary alloc] initWithCapacity:100];
	word_parts = [[NSMutableArray alloc] initWithCapacity:100];
	
	[charList addObject:char01];
	[charList addObject:char02];
	[charList addObject:char03];
	[charList addObject:char04];
	[charList addObject:char05];
	[charList addObject:char06];
	[charList addObject:char07];
	[charList addObject:char08];
	[charList addObject:char09];
	
	[charList addObject:char10];
	[charList addObject:char11];
	[charList addObject:char12];
	[charList addObject:char13];
	[charList addObject:char14];
	[charList addObject:char15];
	[charList addObject:char16];
	[charList addObject:char17];
	[charList addObject:char18];
	[charList addObject:char19];
	
	[charList addObject:char20];
	[charList addObject:char21];
	[charList addObject:char22];
	[charList addObject:char23];
	[charList addObject:char24];
	[charList addObject:char25];
	[charList addObject:char26];
	[charList addObject:char27];
	[charList addObject:char28];
	[charList addObject:char29];
	
	[charList addObject:char30];
	[charList addObject:char31];
	[charList addObject:char32];
	[charList addObject:char33];
	[charList addObject:char34];
	[charList addObject:char35];
	[charList addObject:char36];
	[charList addObject:char37];
	[charList addObject:char38];
	[charList addObject:char39];
	
	[charList addObject:char40];
	[charList addObject:char41];
	[charList addObject:char42];
	[charList addObject:char43];
	[charList addObject:char44];
	[charList addObject:char45];
	[charList addObject:char46];
	[charList addObject:char47];
	[charList addObject:char48];
	[charList addObject:char49];
	
	[charList addObject:char50];
	[charList addObject:char51];
	[charList addObject:char52];
	
	[self toggleChars:FALSE];
	[searchBtn setEnabled:FALSE];
	[wordLoadPi setHidden:TRUE];
	
	wtds = [[WordTableDataSource alloc] init];
	[resultTbl setDataSource:wtds];
	[resultTbl setDelegate:self];
	
	[wordCount setIntValue:0];
	
	// reads in and processes the configuration 
	// file which specifies how the various tiles
	// of the grid are inter-connected.
	//
	{
		NSString * path = [[NSBundle mainBundle] pathForResource:@"config" ofType:@"xml"];
		NSData * data = [[NSFileManager defaultManager] contentsAtPath:path];
		NSDictionary * config_dict = [NSPropertyListSerialization propertyListFromData:data mutabilityOption:kCFPropertyListImmutable format:nil errorDescription:nil];
		NSArray * chars = [config_dict objectForKey:@"chars"];
		
		[self initBlocks:chars];
	}
	
	// pre-load the word-list text field with the 
	// full path to the included word list file,
	// which is in the application bundle.
	//
	[fileTxt setStringValue:[[NSBundle mainBundle] pathForResource:@"2+2lemma" ofType:@"txt"]];
}

/* No idea.
 *
 *
 */
- (id)initWithCoder:(NSCoder *)decoder
{
	return self;
}

/* When the "Load..." button is clicked, this
 * function should load a word list and make
 * it ready for searching.
 *
 * This function detaches a new thread for the
 * purpose of actually processing the word list
 * instead of locking up the UI.
 *
 */
- (IBAction)actionChoose:(id)sender
{
	[self toggleChars:FALSE];
	[searchBtn setEnabled:FALSE];
	[wordLoadPi setHidden:FALSE];
	[wordLoadPi startAnimation:self];
	[loadBtn setEnabled:FALSE];
	[fileTxt setEnabled:FALSE];
	[minWordLen setEnabled:FALSE];
	[maxWordLen setEnabled:FALSE];
	
	[NSThread detachNewThreadSelector:@selector(initWords:) toTarget:self withObject:[fileTxt stringValue]];
}

/* Perform a search. This process completes
 * nearly instantaneously, so there isn't much
 * of a point in detaching a new thread.
 *
 */
- (IBAction)actionStart:(id)sender
{
	[self search];
}

/*
 *
 *
 */
- (void)search
{
	NSLog(@"%s.. starting", __PRETTY_FUNCTION__);
	
	struct bw_block * path[100] = { 0 };
	struct bw_block * block_cur = NULL;
	struct bw_block * block_new = NULL;
	char word[100] = { 0 };
	int depth_cur = 0;
	int grid_pos = 0;
	int word_pos = 0;
	int word_len_max = [maxWordLen intValue];
	int word_len_min = [minWordLen intValue];
	struct timeval time_beg, time_end;
	int skip_word_check = 0;
	
	if (word_len_max == 0)
		word_len_max = 99;
	
	// copy the letters from the ui "grid" into the bw_block[52]
	// array.
	[self readGrid];
	[wtds deleteAll];
	[resultTbl reloadData];
	[matchCount setStringValue:@""];
	
	gettimeofday(&time_beg, NULL);
	
	// for each of the fifty-two letters in the grid...
	for (grid_pos = 0; grid_pos < block_cnt; ++grid_pos) {
		block_cur = path[depth_cur++] = &blocks[grid_pos];
		block_cur->used = 1;
		word[word_pos++] = block_cur->letter;
		
		// until we have exhausted all options for this 
		// particular top-level letter in the grid...
		while (depth_cur > 0) {
			word[word_pos] = '\0';
			
			NSMutableDictionary * wp = [word_parts objectAtIndex:word_pos-1];
			
			// this is the short-circuit code that makes
			// this code so fast. if there are no words
			// of this length (or greater), or if there
			// are no words that start with this current
			// string, then there is no point is continuing
			// on this path. back up one level and move on.
			if (wp == nil || nil == [wp objectForKey:[NSString stringWithUTF8String:word]])
			{
				--depth_cur;
				--word_pos;
				block_cur->used = 0;
				block_cur->link_pos = 0;
				path[depth_cur] = NULL;
				word[word_pos] = '\0';
				block_cur = path[depth_cur-1];
				skip_word_check = 1;
			}
			
			// check to see if there is a word match.
			if (skip_word_check == 0)
				if (word_pos >= word_len_min && [words objectForKey:[NSString stringWithUTF8String:word]])
					[self showWordMatch:(char *)&word path:path];
			
			skip_word_check = 0;
			
			// iterate until we find the next block, which might
			// involve backing up one or more positions in the 
			// stack before we find a place where we can iterate.
			while (depth_cur > 0) {
				// if we're as deep as we're allowed to go, or if
				// there is no way to dig deeper from this point,
				// pop this item off the stack and loop again.
				if (depth_cur >= word_len_max || block_cur->link_pos >= block_cur->link_cnt) {
					--depth_cur;
					--word_pos;
					block_cur->used = 0;
					block_cur->link_pos = 0;
					path[depth_cur] = NULL;
					word[word_pos] = '\0';
					block_cur = path[depth_cur-1];
				}
				// if the next block is one that has already been
				// used, loop again so that we can check the next.
				else if (block_cur->links[block_cur->link_pos]->used) {
					block_cur->link_pos++;
					continue;
				}
				else {
					block_new = path[depth_cur++] = block_cur->links[block_cur->link_pos++];
					block_cur = block_new;
					
					word[word_pos++] = block_cur->letter;
					block_cur->used = 1;
					
					break;
				}
			}
		}
	}
	
	gettimeofday(&time_end, NULL);
	
	[timeTxt setIntValue:(time_end.tv_sec - time_beg.tv_sec)];
	[matchCount setIntValue:[wtds count]];
	
	NSLog(@"%s.. done! (%d seconds)", __PRETTY_FUNCTION__, (time_end.tv_sec - time_beg.tv_sec));
}

/* Add a new word to the list of results.
 *
 */
- (void)showWordMatch:(char *)word path:(struct bw_block **)path
{
	NSMutableString * pathStr = [[NSMutableString alloc] init];
	
	while (*path != NULL) {
		if ((*path)->letter != 0)
			[pathStr appendFormat:@"%d ", (*path)->numb];
		
		path++;
	}
	
	BWWord * bww = [[BWWord alloc] init];
	bww.word = [NSString stringWithUTF8String:word];
	bww.path = pathStr;
	bww.leng = [NSNumber numberWithInt:[bww.word length]];
	
	// calculate the point value for the word based
	// on the character values.
	//
	{
		int pos=0, value=0;
		
		for (pos = 0; pos < strlen(word); ++pos)
			value += charValues[word[pos] - 97];
		
		bww.valu = [NSNumber numberWithInt:value];
	}
	
	// if there's a "q" in the word, we need to
	// insert the "u" following the "q", since we
	// stripped them from the source word list,
	// such that they will now show up properly
	// in the results table.
	//
	{
		NSRange range = [bww.word rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@"q"]];
		
		if (range.location != NSNotFound) {
			if (range.location == 0)
				bww.word = [NSString stringWithFormat:@"qu%@", [bww.word substringFromIndex:range.location+1]];
			else
				bww.word = [NSString stringWithFormat:@"%@qu%@", [bww.word substringToIndex:range.location], [bww.word substringFromIndex:range.location+1]];
		}
	}
	
	[wtds addWord:bww];
	[resultTbl reloadData];
}

/* Not supported.
 *
 */
- (IBAction)actionStop:(id)sender
{
	
}

/* Read all of the letters in the grid into our
 * internal list which we can search.
 *
 */
- (void)readGrid
{
	int i = 0;
	
	for (i = 0; i < [charList count]; ++i)
		blocks[i].letter = [[[charList objectAtIndex:i] stringValue] UTF8String][0];
}

/* Reads in the contents of the file specified in
 * the argument string and separates the individ-
 * ual words based on white space.
 *
 * Each of the words is stored in the "words" 
 * dictionary, while every possible starting sub-
 * string is stored in a dictionary in the "word_
 * parts" array based on the length of the sub-
 * string.
 *
 */
- (void)initWords:(NSString *)file
{
	NSString * words_str = [NSString stringWithContentsOfFile:file encoding:NSASCIIStringEncoding error:nil];
	NSArray * word_list = [words_str componentsSeparatedByCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
	NSString * substr = nil;
	
	int len_min = [minWordLen intValue];
	int len_max = [maxWordLen intValue];
	
	[words removeAllObjects];
	[word_parts removeAllObjects];
	
	for (NSString * word in word_list) {
		if (len_min > 0 && [word length] < len_min)
			continue;
		
		if (len_max > 0 && [word length] > len_max)
			continue;
		
		[words setObject:word forKey:word];
		
		// set up all of the indexes, which include every 
		// starting substring for all of the words.
		{
			int char_pos = 0;
			NSMutableDictionary * wp = nil;
			
			for (char_pos = 1; char_pos <= [word length]; ++char_pos) {
				if (char_pos > [word_parts count])
					[word_parts addObject:(wp = [[NSMutableDictionary alloc] initWithCapacity:100])];
				else
					wp = [word_parts objectAtIndex:char_pos-1];
				
				if (nil == [wp objectForKey:(substr = [word substringToIndex:char_pos])])
					[wp setObject:substr forKey:substr];
			}
		}
		
		if ([words count] % 500 == 0)
			[wordCount setIntValue:[words count]];
	}
	
	[wordCount setIntValue:[words count]];
	
	[self toggleChars:TRUE];
	[searchBtn setEnabled:TRUE];
	[wordLoadPi setHidden:TRUE];
	[wordLoadPi stopAnimation:self];
	[loadBtn setEnabled:TRUE];
	[fileTxt setEnabled:TRUE];
	[minWordLen setEnabled:TRUE];
	[maxWordLen setEnabled:TRUE];
}

/* Allocates the list of "blocks", which are our
 * internal representation of the tiles of the
 * grid.
 *
 * Creates "links" between each of the blocks 
 * using pointers, based on the specifications of
 * the config.xml file.
 *
 */
- (void)initBlocks:(NSArray *)config
{
	NSLog(@"%s.. config.count = %d", __PRETTY_FUNCTION__, [config count]);
	NSLog(@"%s.. %@", __PRETTY_FUNCTION__, config);
	
	int pos_cnt = 0;
	struct bw_block * block_ptr = NULL;
	
	block_cnt = [config count];
	
	if (NULL == (blocks = (struct bw_block *)malloc(sizeof(struct bw_block) * block_cnt))) {
		NSLog(@"%s.. failed to malloc!", __PRETTY_FUNCTION__);
		return;
	}
	
	block_ptr = blocks;
	
	memset(blocks, 0, sizeof(struct bw_block) * block_cnt);
	
	for (NSArray * block_data in config) {
		NSLog(@"%s.. pos_cnt=%d", __PRETTY_FUNCTION__, pos_cnt);
		
		block_ptr->numb = pos_cnt++;
		block_ptr->links = (struct bw_block **)malloc(sizeof(void *) * [block_data count]);
		
		if (block_ptr->links == NULL) {
			NSLog(@"%s.. [2] failed to malloc!", __PRETTY_FUNCTION__);
			return;
		}
		
		NSLog(@"%s.. link_cnt=%d", __PRETTY_FUNCTION__, [block_data count]);
		
		block_ptr->link_cnt = [block_data count];
		memset(block_ptr->links, 0, sizeof(void *) * [block_data count]);
		
		int link_cnt = 0;
		
		for (NSNumber * link in block_data) {
			block_ptr->links[link_cnt++] = &blocks[[link intValue]];
			NSLog(@"%s.. link %d = %d", __PRETTY_FUNCTION__, (link_cnt-1), [link intValue]);
		}
		
		NSLog(@"\n\n");
		block_ptr++;
	}
	
	{
		int i, j;
		
		for (i = 0; i < block_cnt; ++i) {
			NSLog(@"%s.. [%02d] numb=%d, lcnt=%d", __PRETTY_FUNCTION__, i, blocks[i].numb, blocks[i].link_cnt);
			
			for (j = 0; j < blocks[i].link_cnt; ++j)
				NSLog(@"%s..   [%d] numb=%d, lcnt=%d", __PRETTY_FUNCTION__, j, blocks[i].links[j]->numb, blocks[i].links[j]->link_cnt);
		}
	}
}

/* Clear the characters specified in the UI.
 *
 *
 */
- (void)clearChars
{
	for (NSTextField * txtfld in charList)
		[txtfld setStringValue:@""];
}

/* Enable/Disable the grid UI.
 *
 *
 */
- (void)toggleChars:(BOOL)onOff
{
	for (NSTextField * txtfld in charList)
		[txtfld setEnabled:onOff];
}

/* this is for the changing selection in
 * the result table
 *
 */
- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
	int ndx = [resultTbl selectedRow];
	
	// clear any existing highlight
	for (NSTextField * fld in charList)
		[fld setBackgroundColor:txtColor];
	
	if (ndx >= 0) {
		BWWord * word = [wtds wordAtIndex:ndx];
		NSArray * path = [word.path componentsSeparatedByCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
		
		for (NSString * part in path) {
			if (part != nil && [part length] > 0)
				[[charList objectAtIndex:[part intValue]] setBackgroundColor:[NSColor yellowColor]];
		}
	}
}

/* Maybe one day we can use this to auto-tab 
 * after a character is typed into the grid.
 *
 */
- (void)textDidChange:(NSNotification *)aNotification
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
}

@end
